import type { ThemedToken, ThemeRegistration } from '@shikijs/types'
import type { ScrollViewProps, StyleProp, TextStyle, ViewStyle } from 'react-native'
import type { SyntaxHighlighterProps } from 'src/types/shiki'
import transform, { type StyleTuple } from 'css-to-react-native'
import React, { useMemo } from 'react'
import { Platform, ScrollView, StyleSheet, Text, View } from 'react-native'
import { useSyntaxHighlighter } from '../../hooks/useSyntaxHighlighter'

const monospaceFont = Platform.select({
  ios: 'Menlo',
  android: 'monospace',
  default: 'monospace',
})

const baseStyles = StyleSheet.create({
  scrollView: {
    flex: 1,
    minHeight: 20,
  },
  codeContainer: {
    padding: 16,
    minWidth: '100%',
  },
  lineContainer: {
    flexDirection: 'row',
    flexWrap: 'wrap',
  },
  token: {
    fontFamily: monospaceFont,
    fontSize: 14,
  },
})

const ALLOWED_STYLE_PROPERTIES: Record<string, boolean> = {
  color: true,
  background: true,
  backgroundColor: true,
  fontWeight: true,
  fontStyle: true,
}

function cleanStyle(style: Record<string, any>) {
  const styles = Object.entries(style)
    .filter(([key]) => ALLOWED_STYLE_PROPERTIES[key])
    .map<StyleTuple>(([key, value]) => [key, value])
  return transform(styles)
}

export function getRNStylesFromShikiStyle(theme: ThemeRegistration): Record<string, TextStyle> {
  if (!theme?.tokenColors)
    return {}

  const styles: Record<string, TextStyle> = {
    editor: cleanStyle({
      backgroundColor: theme.colors?.['editor.background'] || '#1e1e1e',
      color: theme.colors?.['editor.foreground'] || '#d4d4d4',
    }),
  }

  theme.tokenColors.forEach((token) => {
    if (!token.settings || !token.scope)
      return

    const style = cleanStyle(token.settings)
    const scopes = Array.isArray(token.scope) ? token.scope : [token.scope]

    scopes.forEach((scope) => {
      styles[scope] = style
    })
  })

  return styles
}

interface Props extends SyntaxHighlighterProps {
  scrollViewProps?: ScrollViewProps
  textStyle?: StyleProp<TextStyle>
  containerStyle?: StyleProp<ViewStyle>
}

const SyntaxHighlighter: React.FC<Props> = ({
  text,
  language,
  languageId,
  theme,
  fontSize = 14,
  scrollViewProps,
  textStyle,
  containerStyle,
}) => {
  const { tokens } = useSyntaxHighlighter({
    text,
    language,
    languageId: languageId || String(language),
    theme,
  })

  const themeStyles = useMemo(() => getRNStylesFromShikiStyle(theme as ThemeRegistration), [theme])

  const renderToken = (token: ThemedToken, rowIndex: number, tokenIndex: number) => {
    const tokenStyle = cleanStyle({
      color: token.color || themeStyles.editor?.color,
      fontStyle: token.fontStyle ? 'italic' : 'normal',
      fontWeight: token.fontStyle ? 'bold' : 'normal',
    })

    return (
      <Text
        key={`${rowIndex}-${tokenIndex}`}
        style={[baseStyles.token, tokenStyle, textStyle, { fontSize }]}
      >
        {token.content}
      </Text>
    )
  }

  return (
    <ScrollView
      horizontal
      {...scrollViewProps}
      style={[baseStyles.scrollView, themeStyles.editor, scrollViewProps?.style]}
      contentContainerStyle={[baseStyles.codeContainer, scrollViewProps?.contentContainerStyle, containerStyle]}
    >
      <View onStartShouldSetResponder={() => true}>
        {tokens.map((line, rowIndex) => (
          <View key={rowIndex} style={baseStyles.lineContainer}>
            {line.map((token, tokenIndex) => renderToken(token, rowIndex, tokenIndex))}
          </View>
        ))}
      </View>
    </ScrollView>
  )
}

export default SyntaxHighlighter
